/*
 * Copyright (c) Citrix Systems, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */


#include <stddef.h>
#include <stdlib.h>

#include "xen_internal.h"
#include <xen/api/xen_common.h>
#include <xen/api/xen_gpu_group.h>
#include <xen/api/xen_string_string_map.h>
#include <xen/api/xen_vgpu.h>
#include <xen/api/xen_vm.h>


XEN_FREE(xen_vgpu)
XEN_SET_ALLOC_FREE(xen_vgpu)
XEN_ALLOC(xen_vgpu_record)
XEN_SET_ALLOC_FREE(xen_vgpu_record)
XEN_ALLOC(xen_vgpu_record_opt)
XEN_RECORD_OPT_FREE(xen_vgpu)
XEN_SET_ALLOC_FREE(xen_vgpu_record_opt)


static const struct_member xen_vgpu_record_struct_members[] =
    {
        { .key = "uuid",
          .type = &abstract_type_string,
          .offset = offsetof(xen_vgpu_record, uuid) },
        { .key = "VM",
          .type = &abstract_type_ref,
          .offset = offsetof(xen_vgpu_record, vm) },
        { .key = "GPU_group",
          .type = &abstract_type_ref,
          .offset = offsetof(xen_vgpu_record, gpu_group) },
        { .key = "device",
          .type = &abstract_type_string,
          .offset = offsetof(xen_vgpu_record, device) },
        { .key = "currently_attached",
          .type = &abstract_type_bool,
          .offset = offsetof(xen_vgpu_record, currently_attached) },
        { .key = "other_config",
          .type = &abstract_type_string_string_map,
          .offset = offsetof(xen_vgpu_record, other_config) }
    };

const abstract_type xen_vgpu_record_abstract_type_ =
    {
       .typename = STRUCT,
       .struct_size = sizeof(xen_vgpu_record),
       .member_count =
           sizeof(xen_vgpu_record_struct_members) / sizeof(struct_member),
       .members = xen_vgpu_record_struct_members
    };


void
xen_vgpu_record_free(xen_vgpu_record *record)
{
    if (record == NULL)
    {
        return;
    }
    free(record->handle);
    free(record->uuid);
    xen_vm_record_opt_free(record->vm);
    xen_gpu_group_record_opt_free(record->gpu_group);
    free(record->device);
    xen_string_string_map_free(record->other_config);
    free(record);
}


bool
xen_vgpu_get_record(xen_session *session, xen_vgpu_record **result, xen_vgpu vgpu)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = vgpu }
        };

    abstract_type result_type = xen_vgpu_record_abstract_type_;

    *result = NULL;
    XEN_CALL_("VGPU.get_record");

    if (session->ok)
    {
       (*result)->handle = xen_strdup_((*result)->uuid);
    }

    return session->ok;
}


bool
xen_vgpu_get_by_uuid(xen_session *session, xen_vgpu *result, char *uuid)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = uuid }
        };

    abstract_type result_type = abstract_type_string;

    *result = NULL;
    XEN_CALL_("VGPU.get_by_uuid");
    return session->ok;
}


bool
xen_vgpu_get_vm(xen_session *session, xen_vm *result, xen_vgpu vgpu)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = vgpu }
        };

    abstract_type result_type = abstract_type_string;

    *result = NULL;
    XEN_CALL_("VGPU.get_VM");
    return session->ok;
}


bool
xen_vgpu_get_gpu_group(xen_session *session, xen_gpu_group *result, xen_vgpu vgpu)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = vgpu }
        };

    abstract_type result_type = abstract_type_string;

    *result = NULL;
    XEN_CALL_("VGPU.get_GPU_group");
    return session->ok;
}


bool
xen_vgpu_get_device(xen_session *session, char **result, xen_vgpu vgpu)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = vgpu }
        };

    abstract_type result_type = abstract_type_string;

    *result = NULL;
    XEN_CALL_("VGPU.get_device");
    return session->ok;
}


bool
xen_vgpu_get_currently_attached(xen_session *session, bool *result, xen_vgpu vgpu)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = vgpu }
        };

    abstract_type result_type = abstract_type_bool;

    XEN_CALL_("VGPU.get_currently_attached");
    return session->ok;
}


bool
xen_vgpu_get_other_config(xen_session *session, xen_string_string_map **result, xen_vgpu vgpu)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = vgpu }
        };

    abstract_type result_type = abstract_type_string_string_map;

    *result = NULL;
    XEN_CALL_("VGPU.get_other_config");
    return session->ok;
}


bool
xen_vgpu_set_other_config(xen_session *session, xen_vgpu vgpu, xen_string_string_map *other_config)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = vgpu },
            { .type = &abstract_type_string_string_map,
              .u.set_val = (arbitrary_set *)other_config }
        };

    xen_call_(session, "VGPU.set_other_config", param_values, 2, NULL, NULL);
    return session->ok;
}


bool
xen_vgpu_add_to_other_config(xen_session *session, xen_vgpu vgpu, char *key, char *value)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = vgpu },
            { .type = &abstract_type_string,
              .u.string_val = key },
            { .type = &abstract_type_string,
              .u.string_val = value }
        };

    xen_call_(session, "VGPU.add_to_other_config", param_values, 3, NULL, NULL);
    return session->ok;
}


bool
xen_vgpu_remove_from_other_config(xen_session *session, xen_vgpu vgpu, char *key)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = vgpu },
            { .type = &abstract_type_string,
              .u.string_val = key }
        };

    xen_call_(session, "VGPU.remove_from_other_config", param_values, 2, NULL, NULL);
    return session->ok;
}


bool
xen_vgpu_create(xen_session *session, xen_vgpu *result, xen_vm vm, xen_gpu_group gpu_group, char *device, xen_string_string_map *other_config)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = vm },
            { .type = &abstract_type_string,
              .u.string_val = gpu_group },
            { .type = &abstract_type_string,
              .u.string_val = device },
            { .type = &abstract_type_string_string_map,
              .u.set_val = (arbitrary_set *)other_config }
        };

    abstract_type result_type = abstract_type_string;

    *result = NULL;
    XEN_CALL_("VGPU.create");
    return session->ok;
}


bool
xen_vgpu_destroy(xen_session *session, xen_vgpu self)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = self }
        };

    xen_call_(session, "VGPU.destroy", param_values, 1, NULL, NULL);
    return session->ok;
}


bool
xen_vgpu_get_all(xen_session *session, struct xen_vgpu_set **result)
{

    abstract_type result_type = abstract_type_string_set;

    *result = NULL;
    xen_call_(session, "VGPU.get_all", NULL, 0, &result_type, result);
    return session->ok;
}


bool
xen_vgpu_get_uuid(xen_session *session, char **result, xen_vgpu vgpu)
{
    abstract_value param_values[] =
        {
            { .type = &abstract_type_string,
              .u.string_val = vgpu }
        };

    abstract_type result_type = abstract_type_string;

    *result = NULL;
    XEN_CALL_("VGPU.get_uuid");
    return session->ok;
}
